@@ -3,7 +3,7 @@
 -- Module      :  CppIfdef
 -- Copyright   :  1999-2004 Malcolm Wallace
 -- Licence     :  LGPL
--- 
+--
 -- Maintainer  :  Malcolm Wallace <Malcolm.Wallace@cs.york.ac.uk>
 -- Stability   :  experimental
 -- Portability :  All
@@ -74,9 +74,9 @@ cpp _ _ _ _ _ [] = return []
 cpp p syms path options (Keep ps) (l@('#':x):xs) =
     let ws = words x
         cmd = if null ws then "" else head ws
-        line = tail ws
-        sym  = head (tail ws)
-        rest = tail (tail ws)
+        line = if null ws then [] else tail ws
+        sym  = if null line then "" else head line
+        rest = if null line then [] else tail line
         def = defineMacro options (sym++" "++ maybe "1" id (un rest))
         un v = if null v then Nothing else Just (unwords v)
         keepIf b = if b then Keep (p:ps) else Drop 1 False (p:ps)
@@ -216,21 +216,27 @@ preExpand st =
 -- | Expansion of symbols.
 expandSymOrCall :: SymTab HashDefine -> TextParser String
 expandSymOrCall st =
-  do  sym <- parseSym
+  do sym <- parseSym
+     if sym=="defined" then do arg <- skip parseSym; convert sym [arg]
+                            <|>
+                            do arg <- skip $ parenthesis (do x <- skip parseSym;
+                                                             skip (return x))
+                               convert sym [arg]
+                            <|> convert sym []
+      else
       ( do  args <- parenthesis (commit $ fragment `sepBy` skip (isWord ","))
             args' <- flip mapM args $ \arg->
                          case runParser (preExpand st) arg of
                              (Left msg, _) -> fail msg
                              (Right s, _)  -> return s
             convert sym args'
-        <|>
-        if (sym=="defined") then do arg <- skip parseSym; convert sym [arg]
-                            else convert sym [] )
+        <|> convert sym []
+      )
   where
     fragment = many1 (satisfy (`notElem`",)"))
     convert "defined" [arg] =
       case lookupST arg st of
-        Nothing | all isDigit arg    -> return arg 
+        Nothing | all isDigit arg    -> return arg
         Nothing                      -> return "0"
         Just (a@AntiDefined{})       -> return "0"
         Just (a@SymbolReplacement{}) -> return "1"
@@ -238,14 +244,16 @@ expandSymOrCall st =
     convert sym args =
       case lookupST sym st of
         Nothing  -> if null args then return sym
-                    else fail (disp sym args++" is not a defined macro")
+                    else return "0"
+                 -- else fail (disp sym args++" is not a defined macro")
         Just (a@SymbolReplacement{}) -> do reparse (replacement a)
                                            return ""
         Just (a@MacroExpansion{})    -> do reparse (expandMacro a args False)
                                            return ""
         Just (a@AntiDefined{})       ->
                     if null args then return sym
-                    else fail (disp sym args++" explicitly undefined with -U")
+                    else return "0"
+                 -- else fail (disp sym args++" explicitly undefined with -U")
     disp sym args = let len = length args
                         chars = map (:[]) ['a'..'z']
                     in sym ++ if null args then ""
